use crate::{
    app::EmailUnsubscribeUrlQueryParams,
    cli::{CLISubcommand, Listen, ListenFdSocket, ListenFdSockets},
    config::Config,
    db::DbThread,
    keys::get_or_make_keys,
    models::Keys,
};
use actix_session::{storage::CookieSessionStore, SessionMiddleware};
use actix_web::{cookie::Key, web, App, HttpServer};
use clap::Parser;
use diesel::Connection;
use diesel_migrations::{embed_migrations, EmbeddedMigrations, MigrationHarness};
use eyre::eyre;
use listenfd::ListenFd;
use openidconnect::EndUserEmail;

pub const MIGRATIONS: EmbeddedMigrations = embed_migrations!();

mod app;
mod cli;
mod client;
mod config;
mod db;
mod keys;
mod models;
mod schema;

pub async fn run_server(
    config: Config,
    db: web::Data<DbThread>,
    keys: web::Data<Keys>,
    listen: Listen,
) -> eyre::Result<()> {
    let config = web::Data::new(config);
    let cookie_session_key = Key::generate();
    let mut server = HttpServer::new(move || {
        let cookie_session =
            SessionMiddleware::builder(CookieSessionStore::default(), cookie_session_key.clone())
                .cookie_name("session".into())
                .build();
        App::new()
            .wrap(cookie_session)
            .app_data(config.clone())
            .app_data(db.clone())
            .app_data(keys.clone())
            .configure(app::all_services)
    });
    let Listen {
        listen_tcp,
        listen_unix,
        listen_fd,
    } = listen;
    for sock in listen_tcp.iter() {
        server = server.bind_auto_h2c(sock)?;
    }
    for sock in listen_unix.iter() {
        server = server.bind_uds(sock)?;
    }
    if let Some(ListenFdSockets(sockets)) = listen_fd {
        for sock in sockets.into_iter() {
            server = match sock {
                ListenFdSocket::Tcp(sock) => server.listen_auto_h2c(sock)?,
                ListenFdSocket::Unix(sock) => server.listen_uds(sock)?,
            };
        }
    }
    log::info!("started server");
    server.run().await?;
    Ok(())
}

pub async fn generate_unsubscribe_urls(
    config: Config,
    keys: web::Data<Keys>,
    emails: Vec<String>,
) -> eyre::Result<()> {
    for email in emails {
        let email = EndUserEmail::new(email);
        let url = EmailUnsubscribeUrlQueryParams::new(&email, &keys).make_url(&config);
        println!("{url}");
    }
    Ok(())
}

pub async fn main() -> eyre::Result<()> {
    env_logger::init();
    let mut listenfd = ListenFd::from_env();
    color_eyre::install()?;
    let mut cli = cli::CLI::parse();
    let mut db = diesel::SqliteConnection::establish(&cli.config.sqlite_db)?;
    db.run_pending_migrations(MIGRATIONS)
        .map_err(|e| eyre!(e))?;
    let keys = web::Data::new(get_or_make_keys(&mut db)?);
    let db = web::Data::new(DbThread::new(
        db,
        cli.config.db_thread_channel_capacity.get().into(),
    ));
    cli.resolve(&mut listenfd).await?;
    match cli.subcommand {
        CLISubcommand::Serve { listen } => run_server(cli.config, db, keys, listen).await,
        CLISubcommand::GenerateUnsubscribeUrls { emails } => {
            generate_unsubscribe_urls(cli.config, keys, emails).await
        }
    }
}
